#include "replace_sem.h"

#ifdef OPTIM
void register_in_event (event_t);
void register_in_timer (scheduler_t);
#endif

/*************************************************/
rewrite_t dispach_rewrite[] =
{
      nothing_rewrite         ,
      cooperate_rewrite       ,
      halt_rewrite            ,
      return_rewrite          ,
      sequence_rewrite        ,
      stop_rewrite            ,
      suspend_rewrite         ,
      resume_rewrite          ,
      if_rewrite              ,
      expression_rewrite      ,
      repeat_rewrite          ,
      generate_rewrite        ,
      generate_valued_rewrite ,
      await_rewrite           ,
      limited_await_rewrite   ,
      atom_rewrite            ,
      join_rewrite            ,
      limited_join_rewrite    ,
      run_rewrite             ,
      while_rewrite           ,
      get_value_rewrite       ,
      automaton_rewrite       ,
};

char rewrite (instruction_t inst)
{
   return dispach_rewrite [(int)inst->type] (inst);
}

/*************************************************/
void fire_thread (thread_t thread)
{
   thread->scheduler->current = thread; // set self
   return_processing (thread,rewrite (thread->run)); 
}

/*************************************************/
char nothing_rewrite (instruction_t inst)
{
   return _TERM;
}

/*************************************************/
char cooperate_rewrite (instruction_t inst)
{
   cooperate_t coop = (cooperate_t) inst;
   if (coop->state) return _TERM;
   coop->state = 1;
   return _COOP;
}

/*************************************************/
char halt_rewrite (instruction_t inst)
{
   return _COOP;
}

/*************************************************/
char return_rewrite (instruction_t inst)
{
   broadcast (self ()->signal);
   return _RTRN;
}

/*************************************************/
char sequence_rewrite (instruction_t inst)
{
   sequence_t s = (sequence_t)inst;

   if (0 == s->state) {   
      int res = rewrite (s->left);
      if (res != _TERM) return res;
      s->state = 1;
   }

   return rewrite(s->right);
}

/***************************************************/
char stop_rewrite (instruction_t inst)
{
   stop_t s = (stop_t)inst;
   thread_t th = evaluate (s->exp);
   stop (th);
   return _TERM;
}

/***************************************************/
char suspend_rewrite (instruction_t inst)
{
   suspend_t s = (suspend_t)inst;
   thread_t th = evaluate (s->exp);
   suspend (th);
   return _TERM;
}

/***************************************************/
char resume_rewrite (instruction_t inst)
{
   resume_t res = (resume_t)inst;
   thread_t th = evaluate (res->exp);
   resume (th);
   return _TERM;
}


/***************************************************/
char if_rewrite (instruction_t inst)
{
   if_t test = (if_t)inst;
   if (0 == test->state) {
      test->state = ((int)evaluate (test->exp)) ? 1 : 2;
   }

   if (1 == test->state) return rewrite (test->left);
   else return rewrite (test->right);   
}

/***************************************************/
char expression_rewrite (instruction_t inst)
{
   inst_from_exp_t i = (inst_from_exp_t)inst;
   evaluate (i->exp);
   return _TERM;
}

/***************************************************/
char repeat_rewrite (instruction_t inst)
{
   repeat_t r = (repeat_t)inst;
   char b;
   
   if (0 == r->state) {
      r->state = 1;
      r->counter = (int)evaluate (r->exp);
   }
   while (1) {
      if (r->counter <= 0) return _TERM;
      b = rewrite (r->body);
      if (b != _TERM) return b;
      r->counter--;
      reset_instruction (r->body);
   }
}

/***************************************************/
char generate_rewrite (instruction_t inst)
{
   generate_t gen = (generate_t)inst;
   event_t event = evaluate (gen->exp);
   generate (event);
   return _TERM;
}

/***************************************************/
char generate_valued_rewrite (instruction_t inst)
{
   generate_valued_t gen = (generate_valued_t)inst;
   event_t event = evaluate (gen->evt_exp);
   void *val = evaluate (gen->val_exp);
   generate_value (event,val);
   return _TERM;
}

/***************************************************/
char await_rewrite (instruction_t inst)
{
   await_t a = (await_t)inst;

   if (0 == a->state) {
      a->event = evaluate (a->exp);
      a->state = 1;
   }
   // the event scheduler must be the current scheduler
   if (is_present (a->event)) return _TERM;
#ifdef OPTIM
   register_in_event (a->event);
   return _WAIT;
#else
   return current_scheduler ()->eoi ? _COOP : _CONT;
#endif
}

/***************************************************/
#ifdef OPTIM
char limited_await_rewrite (instruction_t inst)
{
   limited_await_t a = (limited_await_t)inst;
   scheduler_t sched = current_scheduler ();
   
   if (0 == a->state) {
      a->event = evaluate (a->evt_exp);
      a->limit = (int)evaluate (a->limit_exp);
      self ()->deadline = a->limit;
      a->state = 1;
   }
   // the event scheduler must be the current scheduler
   if (self ()->deadline <= 0) {
      self ()->code = ETIMEOUT;
      return _TERM;
   }
   if (is_present (a->event)) {
      self ()->code = OK;
      return _TERM;
   }
   register_in_event (a->event);
   register_in_timer (sched);
   return _WAIT;
}
#else
char limited_await_rewrite (instruction_t inst)
{
   limited_await_t a = (limited_await_t)inst;
   scheduler_t sched = current_scheduler ();
   
   if (0 == a->state) {
      a->event = evaluate (a->evt_exp);
      a->limit = (int)evaluate (a->limit_exp);
      a->state = 1;
   }
   // the event scheduler must be the current scheduler
   if (a->limit <= 0) {
      self ()->code = ETIMEOUT;
      return _TERM;
   }
   if (is_present (a->event)) {
      self ()->code = OK;
      return _TERM;
   }
   if (sched->eoi) {
      a->limit--;
      return _COOP;
   }
   return _CONT;
}
#endif

/***************************************************/
char atom_rewrite (instruction_t inst)
{
   atom_t a = (atom_t)inst;
   a->selector (a->n,self ()->locals);
   return _TERM;
}

/***************************************************/
char join_rewrite (instruction_t inst)
{
   join_t j = (join_t)inst;

   if (0 == j->state) {
      j->thread = evaluate (j->exp);
      j->state = 1;
   }
   // the thread scheduler must be the current scheduler   
   if ( j->thread->state == _TERM ||
        is_present (j->thread->signal) ) return _TERM;
#ifdef OPTIM
   register_in_event (j->thread->signal);
   return _WAIT;
#else   
   return current_scheduler ()->eoi ? _COOP : _CONT;
#endif
}

/***************************************************/
#ifdef OPTIM
char limited_join_rewrite (instruction_t inst)
{
   limited_join_t j = (limited_join_t)inst;
   scheduler_t sched = current_scheduler ();
   if (0 == j->state) {
      j->thread = evaluate (j->th_exp);
      j->limit = (int)evaluate (j->limit_exp);
      self ()->deadline = j->limit;      
      j->state = 1;
   }
   if (j->thread->state == _TERM) {
      self ()->code = OK;
      return _TERM;
   }
   if (self ()->deadline <= 0) {
      self ()->code = ETIMEOUT;
      return _TERM;
   }
   if (is_present (j->thread->signal)) {
      self ()->code = OK;
      return _TERM;
   }
   register_in_event (j->thread->signal);
   register_in_timer (sched);
   return _WAIT;
}
#else
char limited_join_rewrite (instruction_t inst)
{
   limited_join_t j = (limited_join_t)inst;
   if (0 == j->state) {
      j->thread = evaluate (j->th_exp);
      j->limit  = (int)evaluate (j->limit_exp);
      j->state  = 1;
   }
   if (j->thread->state == _TERM) {
      self ()->code = OK;
      return _TERM;
   }
   if (j->limit <= 0) {
      self ()->code = ETIMEOUT;
      return _TERM;
   }
   if (is_present (j->thread->signal)) {
      self ()->code = OK;
      return _TERM;
   }
   if (current_scheduler ()->eoi) {
      j->limit--;
      return _COOP;
   }
   return _CONT;
}
#endif

/***************************************************/
char run_rewrite (instruction_t inst)
{
   run_t r = (run_t)inst;
   
   if (0 == r->state) {
      r->thread = evaluate (r->exp);
      self ()->called = r->thread;
      r->state = 1;
   }

   if (r->thread->state == _TERM) return _TERM;
   if (is_present (r->thread->signal)) return _TERM;
#ifdef OPTIM
   register_in_event (r->thread->signal);
   return _WAIT;
#else
   return current_scheduler()->eoi ? _COOP : _CONT;
#endif
}

/***************************************************/
char while_rewrite (instruction_t inst)
{
   while_t w = (while_t)inst;
   char r;
   while (1) {
      if (0 == w->state) {
	 int b = (int)evaluate (w->exp);
	 if (!b) return _TERM;
	 w->state = 1;
      }
      r = rewrite (w->body);
      if (r != _TERM) return r;
      reset_instruction (w->body);
      w->state = 0;
   }
}

/***************************************************/
char get_value_rewrite (instruction_t inst)
{
   get_value_t g = (get_value_t)inst;
   int len;

   if (2 == g->state) return _TERM;
   
   if (0 == g->state) {
      g->event = evaluate (g->evt_exp);
      g->num = (int)evaluate (g->num_exp);
      g->state = 1;
   }

   len = event_values_length (g->event);

   if (len > g->num) {
      (*g->var) = nth_of_event_values (g->event,g->num);
      self ()->code = OK;      
      return _TERM;
   }
   if (current_scheduler()->eoi) {
      self ()->code = ENEXT;
      g->state = 2;
      return _COOP;
   }
   
   return _CONT;
}

/***************************************************/
char automaton_rewrite (instruction_t inst)
{
   automaton_t a = (automaton_t)inst;
   return a->fun (self ()->locals);
}
